from south.creator import changes, actions
from south.migration.base import Migrations

from ._utils import ArmAccessTestCase as TestCase

class BasicMigrationsAreRunnableTestCase(TestCase):
    def test_verify_invalid_code_is_not_created(self):
        m = Migrations("arm_access_support", force_creation=False)
        c = changes.InitialChanges(m)
        params = c.get_changes().next()[1]
        actor = actions.AddModel(**params)
        code = actor.forwards_code()
        self.assertNotRegexpMatches(code, r"to=orm\['arm_access.AccessObject'\]")



import datetime

from django.contrib.auth.models import User, AnonymousUser
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.template.response import HttpResponse, TemplateResponse
from django.test import TestCase
from django.utils.unittest import skip

from ._utils import *
from ..models import *
from ..paywalls.base import *
from ..paywalls.subscription import SubscriptionPaywall, SubscriptionChecker
from .arm_access_support.models import *


class DummyRequest(object):
    def __init__(self, user):
        self.user = user


class DummyRedirect(object):
    status_code = 301


def dummy_content_view(content_object):
    def inner(request):
        return TemplateResponse(request, 'content_detail.html', {
            'content_object': content_object,
        })
    return inner


class SubscriptionTestCase(ArmAccessTestCase):
    def setUp(self):
        super(SubscriptionTestCase, self).setUp()
        self.user = User.objects.create_user("bob",
                                             "bob@example.com", "secret")
        self.protected = Level.objects.create(name='Level 1',
                                              is_protected=True)
        self.public = Level.objects.create(name='Level 2', is_protected=False)
        self.now = datetime.datetime.now()
        self.dummy_content = ArmAccessSupportContent.objects.create()
        self.dummy_content.access = []
        self.dummy_content.access.create(
            level=self.protected, start_date=datetime.datetime.now())

        self.dummy_view = dummy_content_view(self.dummy_content)

    def test_subscription_paywall_denies_content_by_default(self):
        paywall = SubscriptionPaywall()
        protected_view = paywall.protect(self.dummy_view,
                template_object_name='content_object')
        request = DummyRequest(self.user)
        self.assertRaises(PermissionDenied, protected_view, request)

    def test_subscription_paywall_can_change_template(self):
        paywall = SubscriptionPaywall(render_on_deny('paywall.html'))
        protected_view = paywall.protect(self.dummy_view,
                template_object_name='content_object')
        response = protected_view(DummyRequest(self.user))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.is_rendered, False)
        self.assertEqual(response.template_name, 'paywall.html')

    def test_subscription_paywall_can_redirect(self):
        paywall = SubscriptionPaywall(redirect_on_deny('/login/'))
        protected_view = paywall.protect(self.dummy_view,
                template_object_name='content_object')
        response = protected_view(DummyRequest(self.user))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/login/')

    def test_subscription_paywall_allows_redirects_through(self):
        paywall = SubscriptionPaywall(redirect_on_deny('/login/'))
        protected_view = paywall.protect(lambda req: DummyRedirect())
        response = protected_view(DummyRequest(self.user))
        self.assertEqual(response.status_code, 301)

    def test_access_granted_by_current_subscription(self):
        paywall = SubscriptionPaywall()
        protected_view = paywall.protect(self.dummy_view,
                template_object_name='content_object')
        request = DummyRequest(self.user)
        self.assertRaises(PermissionDenied, protected_view, request)
        today = datetime.date.today()
        year_end = today + datetime.timedelta(days=365)
        AccessMembership.objects.create(user=self.user, level=self.protected,
                                    start_date=today, end_date=year_end)
        response = protected_view(request)
        AccessMembership.objects.filter(user=self.user).update(active=False)
        request.user = User.objects.get(pk=self.user.pk)
        self.assertRaises(PermissionDenied, protected_view, request)

    def test_non_template_response_raises_improper_response(self):
        def well_intentioned_view(request):
            return HttpResponse('Protect me!')

        paywall = SubscriptionPaywall()
        protected_view = paywall.protect(well_intentioned_view,
                template_object_name='object')
        request = DummyRequest(self.user)
        self.assertRaises(ImproperResponse, protected_view, request)

    def test_content_only_in_protected_pub_is_marked_as_protected(self):
        c = ArmAccessSupportContent.objects.create()

        #no publication = not protected
        self.assertFalse(SubscriptionChecker.required_permissions(c)
                is not None)
        c.access = Assignment(level=self.protected,
                              start_date=datetime.datetime.now())
        self.assertTrue(SubscriptionChecker.required_permissions(c)
                is not None)

    def test_content_only_in_unprotected_pub_is_not_protected(self):
        c = ArmAccessSupportContent.objects.create()
        c.access = [Assignment(level=self.protected,
                               start_date=datetime.datetime.now()),
                    Assignment(level=self.public,
                               start_date=datetime.datetime.now()),
                   ]
        self.assertFalse(SubscriptionChecker.required_permissions(c)
                is not None)

    def test_content_published_in_free_pub_is_not_protected(self):
        c = ArmAccessSupportContent.objects.create()

        c.access = []
        c.access.create(level=self.protected,
                        start_date=datetime.datetime.now())
        self.assertTrue(SubscriptionChecker.required_permissions(c)
                is not None)
        c.access.create(level=self.public, start_date=datetime.datetime.now())
        self.assertFalse(SubscriptionChecker.required_permissions(c)
                is not None)
        #sanity check that we added two publications in this test
        self.assertEqual(c.access.assignments.count(), 2)

    def test_content_unpublished_in_free_pub_is_still_protected(self):
        c = ArmAccessSupportContent.objects.create()

        c.access = Assignment(level=self.protected,
                              start_date=datetime.datetime.now())
        self.assertTrue(SubscriptionChecker.required_permissions(c)
                is not None)
        c.access.add(Assignment(level=self.public,
                                start_date=datetime.datetime.now() +
                                           datetime.timedelta(days=999)))
        #since it's not published in public, it's still protected
        self.assertTrue(SubscriptionChecker.required_permissions(c)
                is not None)

    def test_anonymous_user_has_access_to_unprotected_content(self):
        c = ArmAccessSupportContent.objects.create()
        self.assertTrue(SubscriptionChecker.has_permission(
                        DummyRequest(AnonymousUser()), c))

    def test_anonymous_user_has_access_to_content_with_unprotected(self):
        c = ArmAccessSupportContent.objects.create()
        c.access = Assignment(level=self.public, start_date=self.now)
        self.assertTrue(SubscriptionChecker.has_permission(
            DummyRequest(AnonymousUser()), c))

    def test_anonymous_user_has_no_access_to_protected_content(self):
        c = ArmAccessSupportContent.objects.create()
        c.access = Assignment(level=self.protected, start_date=self.now)
        self.assertFalse(SubscriptionChecker.has_permission(
                        DummyRequest(AnonymousUser()), c))

    def test_staff_user_has_access_to_protected_content(self):
        c = ArmAccessSupportContent.objects.create()
        c.access = Assignment(level=self.protected, start_date=self.now)
        self.user.is_staff = True
        self.assertTrue(SubscriptionChecker.has_permission(
                                            DummyRequest(self.user), c))

    def test_user_with_unrelated_access_cant_access_protected_content(self):
        c = ArmAccessSupportContent.objects.create()
        c.access = Assignment(level=self.protected, start_date=self.now)
        new_level = Level.objects.create(name="Magic Pony")
        today = datetime.date.today()
        year_end = today + datetime.timedelta(days=365)
        AccessMembership.objects.create(user=self.user, level=new_level,
                                    start_date=today, end_date=year_end)
        self.assertFalse(SubscriptionChecker.has_permission(
                                        DummyRequest(self.user), c))



from armstrong.dev.tests.utils.users import generate_random_user

from ._utils import *
from ..models import *
from .arm_access_support.models import *


class AccessModelsTestCase(ArmAccessTestCase):
    def testAccessMembershipStr(self):
        args = {}
        args['user'] = generate_random_user()
        args['level'] = Level(name="foo", is_protected=False)
        args['start_date'] = datetime.datetime.now()
        args['end_date'] = datetime.datetime.now()
        self.assertEqual('%s (foo)' % (args['user']),
                         str(AccessMembership(**args)))




from django.db import models
from django.db.models.fields.related \
        import ReverseSingleRelatedObjectDescriptor
from .models import AccessObject
from .widgets import AccessWidget
from .forms import AccessFormField


class AccessField(models.OneToOneField):
    def __init__(self):
        super(AccessField, self).__init__(AccessObject, to_field=None,
                null=True, blank=True)

    def contribute_to_class(self, cls, name):
        super(AccessField, self).contribute_to_class(cls, name)
        setattr(cls, self.name, AccessDescriptor(self))

    def formfield(self, **kwargs):
        return AccessFormField(**kwargs)

    def south_field_triple(self):
        return ('armstrong.core.arm_access.fields.AccessField', [], {})


class AccessDescriptor(ReverseSingleRelatedObjectDescriptor):
    def __set__(self, instance, value):
        # we accept None, a list of Assignments or a single assignment
        obj = self.__get__(instance)
        if obj is None and value is not None:
            obj = AccessObject.objects.create()
            instance.access_id = obj.id

        if value is None:
            if obj is not None:
                instance.access_id = None
                instance.save()
                obj.delete()
            # use our super class for cache invalidation
            super(AccessDescriptor, self).__set__(instance, None)
        else:
            obj.clear()
            if hasattr(value, '__iter__'):
                obj.add(*value)
            else:
                obj.add(value)
            super(AccessDescriptor, self).__set__(instance, obj)



import datetime

from django.conf import settings
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q
from django.db.models.query import QuerySet
from django.utils.translation import ugettext_lazy as _


class Level(models.Model):
    """
    Represents a discrete access level that can be granted to a user
    """
    name = models.CharField(_('name'), max_length=50, unique=True)
    is_protected = models.BooleanField()

    class Meta:
        verbose_name = _('access level')
        verbose_name_plural = _('access levels')
        ordering = ('name',)

    def __unicode__(self):
        return self.name


class AccessObject(models.Model):
    """
    An access model represents a set of access assignments. A content object
    that needs to be access restricted should have relate to exactly one
    AccessModel
    """
    @property
    def current_assignments(self):
        now = datetime.datetime.now()
        return self.assignments.filter(end_date__gte=now,
                                       start_date__lte=now).select_related()

    def add(self, *args, **kwargs):
        return self.assignments.add(*args, **kwargs)

    def create(self, *args, **kwargs):
        return self.assignments.create(*args, **kwargs)

    def clear(self):
        return self.assignments.all().delete()

    def __unicode__(self):
        return u'AccessObject - %i' % self.id

    def is_protected(self):
        for assignment in self.current_assignments:
            if not assignment.level.is_protected:
                return False
        return True



class Assignment(models.Model):
    """
    An access assignment grants access to an AccessModel for a set period of
    time
    """
    access_object = models.ForeignKey(AccessObject, related_name='assignments')
    level = models.ForeignKey(Level,
                              verbose_name=_("access level"),
                              related_name='assignments')
    start_date = models.DateTimeField(_('start date'))
    end_date = models.DateTimeField(_('end date'),
            default=datetime.datetime.max)

    def __unicode__(self):
        return u"%s - %s" % (self.level, self.access_object)


class AccessMembershipQuerySet(QuerySet):
    def current(self):
        today = datetime.date.today()
        return self.filter(active=True,
                           start_date__lte=today, end_date__gte=today)


class AccessMembershipManager(models.Manager):
    """
    Manager that provides current and active access level memberships.
    """
    use_for_related_fields = True

    def get_query_set(self):
        return AccessMembershipQuerySet(self.model)

    def current(self):
        return self.get_query_set().current()


class AccessMembership(models.Model):
    """
    A user subcription to a particular access level.
    """
    objects = AccessMembershipManager()

    user = models.ForeignKey('auth.User',
                             related_name='access_memberships')
    level = models.ForeignKey(Level, related_name='access_memberships')
    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    active = models.BooleanField(default=True)

    def __unicode__(self):
        return "%s (%s)" % (self.user, self.level)

    @property
    def remaining(self):
        now = datetime.datetime.now()
        return self.end_date - now



import datetime
import sys

from django.db.models import ObjectDoesNotExist
from ..models import Assignment
from .base import protect, redirect_on_deny, render_on_deny, raise_on_deny


class SubscriptionPaywall(object):
    """
    This content authorization backend is used to restrict access to content
    based on a user's subscription to the content's access level.

    Subscriptions will only be checked for content which is published
    exclusively in an access level with subscription_required set to True

    This paywall requires a TemplateResponse.

    :param permission_denied: a method that takes in a response for a content
        object that has been denied and returns a response appropriate for
        denied content
    """

    def __init__(self, permission_denied=None):
        if permission_denied is not None:
            self.permission_denied = permission_denied
        else:
            self.permission_denied = raise_on_deny

    def protect(self, view, permission_denied=None, **kwargs):
        if permission_denied is None:
            permission_denied = self.permission_denied
        return protect(view,
                       SubscriptionChecker.has_permission,
                       permission_denied,
                       **kwargs)


class SubscriptionChecker:
    @classmethod
    def required_permissions(cls, content_object):
        if content_object.access is None:
            return None
        assignments = content_object.access.current_assignments
        permissions = []
        for assignment in assignments:
            if assignment.level.is_protected:
                permissions.append(assignment.level)
            else:
                # there is an unprotected level with access
                return None
        return permissions

    @classmethod
    def has_permission(cls, request, content_object):
        required = SubscriptionChecker.required_permissions(content_object)

        # Only check content with subscription required
        if required is None:
            return True

        # Anonymous users can't view subscription content
        if not request.user.is_authenticated():
            return False

        # All staff can view subscription content
        if request.user.is_staff:
            return True

        # Also users with an active subscription to this level
        try:
            today = datetime.date.today()
            memberships = request.user.access_memberships.current()
            for membership in memberships:
                if membership.level in required:
                    return True
            return False
        except ObjectDoesNotExist:
            return False



from functools import wraps
from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse


class ImproperResponse(Exception):
    "An unexpected response type was returned from the view"
    pass


def redirect_on_deny(redirect_uri):
    def redirect(response):
        redirect = HttpResponseRedirect(redirect_uri)
        response.status_code = redirect.status_code
        response['Location'] = redirect['Location']
        return response
    return redirect


def render_on_deny(template_name):
    def rerender(response):
        response.template_name = template_name
        return response
    return rerender


def raise_on_deny(response):
    raise PermissionDenied


def protect(view,
            permission_check,
            permission_denied=raise_on_deny,
            template_object_name=None):
    @wraps(view)
    def wrapped(request, *args, **kwargs):
        response = view(request, *args, **kwargs)

        if response.status_code != 200:
            # Redirects, errors, etc. should pass through.
            return response

        # Can't handle non-template responses
        if not isinstance(response, TemplateResponse):
            raise ImproperResponse('Protected view must return'
                                   ' a TemplateResponse object')

        if template_object_name is None:
            content_object = view.get_object()
        else:
            content_object = response.context_data[template_object_name]

        if not permission_check(request, content_object):
            return permission_denied(response)
        else:
            return response

    return wrapped



from django.db import models
from ...models import AccessObject
from ...mixins import AccessMixin


class ArmAccessSupportContent(AccessMixin, models.Model):
    name = models.CharField(max_length=255)



from django.forms import Widget, Form, SplitDateTimeField
from django.forms.models import modelformset_factory, ModelForm
from django.template import loader, Context
from django.contrib.admin.widgets import AdminSplitDateTime
from .models import Assignment
from datetime import datetime
from functools import partial
from itertools import count
import re


class InlineAssignmentForm(ModelForm):
    start_date = SplitDateTimeField(widget=AdminSplitDateTime)
    end_date = SplitDateTimeField(widget=AdminSplitDateTime,
                                  required=False)

    def __init__(self, *args, **kwargs):
        super(InlineAssignmentForm, self).__init__(*args, **kwargs)
        # necessary because SplitDateTimeWidget doesn't follow the spec and
        # call callables, instead it tries to unpack the value
        if self.initial is not None and not 'start_date' in self.initial:
            self.initial['start_date'] = datetime.now()

    class Meta:
        exclude = ('access_object', 'id')

    def clean(self):
        data = super(InlineAssignmentForm, self).clean()
        if data['end_date'] is None:
            data['end_date'] = datetime.max
        return data

    # we want to always create new Assignments because we create a new access
    # object
    def has_changed(self):
        return True


EmptyAssignmentFormSet = modelformset_factory(Assignment,
                                         form=InlineAssignmentForm,
                                         extra=1,
                                        )


AssignmentFormSet = modelformset_factory(Assignment,
                                         form=InlineAssignmentForm,
                                         extra=0,
                                        )


class AccessWidget(Widget):
    template_name = 'arm_access/access_object_widget.html'

    # pulled from https://github.com/brutasse/django/blob/15667-template-widgets/django/forms/widgets.py
    # hoping that it becomes the new django widget API
    def render(self, name, value, attrs=None):
        context = self.get_context(name, value, attrs=attrs)
        return loader.render_to_string(self.template_name, context)

    def get_context(self, name, value, attrs=None):
        prefix = name + '-assignments'
        if value is None:
            # if object.access is None
            assignments = EmptyAssignmentFormSet(prefix=prefix,
                    queryset=Assignment.objects.none())
        elif type(value) in (int, long):
            # initial code path, we get the ID of the access object
            assignments = AssignmentFormSet(prefix=prefix,
                    queryset=Assignment.objects.filter(access_object=value))
        else:
            # if form submit failed with validation errors, we are passed an
            # AssignmentFormSet, so we just want to pass it through
            assignments = value

        return Context({
            'hidden': self.is_hidden,
            'name': name,
            'required': self.is_required,
            'is_public': value is None,
            'assignments': assignments
        })

    def value_from_datadict(self, data, files, name):
        if name + '_is_public' in data:
            return None
        return AssignmentFormSet(data, prefix=name + '-assignments')

    # We need a choices property because the django admin decides to wrap us in
    # a RelatedFieldWidgetWrapper if our field inherits from ForeignKey, which
    # AccessField does
    def _set_choices(self, value):
        return

    def _get_choices(self):
        return []

    choices = property(_get_choices, _set_choices)



import datetime

from django.contrib.auth.models import User

from ..models import Level, AccessMembership
from ._utils import *


class TestAccessMemberships(ArmAccessTestCase):
    def setUp(self):
        self.user1 = User.objects.create_user("bob",
                                              "bob@example.com", "secret")
        self.user2 = User.objects.create_user("jim",
                                              "jim@example.com", "secret")
        self.al1 = Level.objects.create(name='Level 1')
        self.al2 = Level.objects.create(name='Level 2')

    def test_model(self):
        self.assertEqual(AccessMembership.objects.count(), 0)
        today = datetime.date.today()
        year_end = today + datetime.timedelta(days=365)
        s = AccessMembership(user=self.user1, level=self.al1,
                         start_date=today, end_date=year_end)
        s.save()
        self.assertEqual(s.active, True)
        self.assertEqual(AccessMembership.objects.count(), 1)
        self.assertEqual(self.al1.access_memberships.count(), 1)
        self.assertEqual(self.user1.access_memberships.count(), 1)

    def test_current_manager(self):
        self.assertEqual(AccessMembership.objects.count(), 0)
        self.assertEqual(AccessMembership.objects.current().count(), 0)
        today = datetime.date.today()
        year_end = today + datetime.timedelta(days=365)
        s = AccessMembership(user=self.user1, level=self.al1,
                         start_date=today, end_date=year_end)
        s.save()
        self.assertEqual(AccessMembership.objects.count(), 1)
        self.assertEqual(AccessMembership.objects.current().count(), 1)
        s.start_date = today + datetime.timedelta(days=1)
        s.save()
        self.assertEqual(AccessMembership.objects.count(), 1)
        self.assertEqual(AccessMembership.objects.current().count(), 0)
        s.start_date = s.end_date = today - datetime.timedelta(days=1)
        s.save()
        self.assertEqual(AccessMembership.objects.count(), 1)
        self.assertEqual(AccessMembership.objects.current().count(), 0)
        s.end_date = today + datetime.timedelta(days=1)
        s.save()
        self.assertEqual(AccessMembership.objects.count(), 1)
        self.assertEqual(AccessMembership.objects.current().count(), 1)
        s.active = False
        s.save()
        self.assertEqual(AccessMembership.objects.count(), 1)
        self.assertEqual(AccessMembership.objects.current().count(), 0)

    def test_current_manager_is_available_through_user(self):
        self.assertEqual(self.user1.access_memberships.count(), 0)
        today = datetime.date.today()
        year_end = today + datetime.timedelta(days=365)
        s = AccessMembership(user=self.user1, level=self.al1,
                         start_date=today, end_date=year_end)
        s.save()
        self.assertEqual(self.user1.access_memberships.count(), 1)
        self.assertEqual(
                self.user1.access_memberships.current().count(), 1)

    # Regression
    def test_users_have_different_current_query_sets(self):
        today = datetime.date.today()
        year_end = today + datetime.timedelta(days=365)
        user1_access_levels = self.user1.access_memberships.current()
        user2_access_levels = self.user2.access_memberships.current()
        self.assertEqual(user1_access_levels.count(), 0)
        self.assertEqual(user2_access_levels.count(), 0)
        s1 = AccessMembership(user=self.user1, level=self.al1,
                          start_date=today, end_date=year_end)
        s1.save()
        self.assertEqual(user1_access_levels.count(), 1)
        self.assertEqual(user2_access_levels.count(), 0)
        s2 = AccessMembership(user=self.user2, level=self.al1,
                          start_date=today, end_date=year_end)
        s2.save()
        self.assertEqual(user1_access_levels.count(), 1)
        self.assertEqual(user2_access_levels.count(), 1)
